import pytest
import importlib
import utils as u
import time
from conf_schema import create_user_blogs, fill_user_blogs, drop_user_blogs

try:
    from pytest_mysql import factories

    @pytest.fixture
    def my_dsn(request):
        """Return connection parameters suitable to target driver."""
        is_detached = request.config.getoption("mysql_detached")
        driver = request.config.getoption("mysql_driver")
        database = request.config.getoption("mysql_dbname") or "test"
        if is_detached:
            mp = request.getfixturevalue("mysql_noproc")
            dsn = {
                "password": request.config.getoption("mysql_passwd"),
            }
        else:
            if not u.has_cmd("mysqld"):
                pytest.skip("test needs mysqld")
            mp = request.getfixturevalue("mysql_proc")
            # u.log.debug(f"mp: {mp}")
            # this fixture creates the database as a side effect
            conn = request.getfixturevalue("mysql")
            # u.log.debug(f"conn: {conn}")
            assert mp.running()
            assert mp.host == "localhost"
            # dsn = {"database": "test"} # if driver == "mysql.connector" else {}
            # NOTE mysql complains about host even if a unix socket is used:-/
            dsn = {"unix_socket": mp.unixsocket}
        # add common connection parameters… although host may be unused
        dsn.update(user=mp.user, host=mp.host, port=mp.port, database=database)
        # FIXME passwd?
        # NOTE pip install mysql-connector-python
        # TODO check for mysql.connector version ?
        if driver == "mysql.connector":
            dsn.update(auth_plugin="mysql_native_password")
        u.log.debug(f"my_dsn for {driver}: {dsn}")
        yield dsn

    @pytest.fixture
    def my_conn(request, my_dsn):
        """Return a connection using the expected driver."""
        driver = request.config.getoption("mysql_driver")
        tries = request.config.getoption("mysql_tries")
        db = importlib.import_module(driver)
        fails = 0
        while tries > 0:
            tries -= 1
            try:
                with db.connect(**my_dsn) as conn:
                    tries = 0
                    yield conn
            except Exception as e:
                fails += 1
                u.log.warning(f"{driver} connection failed ({fails}): {e}")
                time.sleep(1.0)

    @pytest.fixture
    def my_db(my_conn):
        """Build the test database."""
        # initial contents
        with my_conn.cursor() as cur:
            for ct in create_user_blogs("mysql"):
                cur.execute(ct)
            my_conn.commit()
            fill_user_blogs(cur, "mysql")
            my_conn.commit()
        # connection to use
        yield my_conn
        # cleanup
        with my_conn.cursor() as cur:
            for dt in drop_user_blogs("mysql"):
                cur.execute(dt)
            my_conn.commit()

except ModuleNotFoundError:
    # provide empty fixtures to please pytest "parsing"

    @pytest.fixture
    def my_dsn():
        raise Exception("undefined fixture")

    @pytest.fixture
    def my_conn():
        raise Exception("undefined fixture")

    @pytest.fixture
    def my_db():
        raise Exception("undefined fixture")



import shutil
import importlib
import logging

log = logging.getLogger("pytest-aiosql")
logging.basicConfig(level=logging.INFO)


def has_cmd(cmd):
    return shutil.which(cmd) is not None


def has_pkg(pkg):
    try:
        importlib.import_module(pkg)
        return True
    except ModuleNotFoundError:
        return False


def has_service(host="localhost", port=22, retry=1):
    import socket
    import time

    while retry > 0:
        retry -= 1
        try:
            tcp_ip = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            tcp_ip.settimeout(1)
            res = tcp_ip.connect_ex((host, port))
            if res == 0:
                return True
            if retry > 0:
                time.sleep(3)
        except Exception as e:
            log.info(f"connection to {(host, port)} failed: {e}")
            if retry > 0:
                time.sleep(3)
        finally:
            tcp_ip.close()
    return False



import inspect
from pathlib import Path
from unittest import mock
import re

import aiosql
from aiosql import SQLParseException, SQLLoadException
from aiosql.queries import Queries
from aiosql.query_loader import QueryLoader

import pytest

pytestmark = [
    pytest.mark.misc,
]


@pytest.fixture
def sql_dir():
    return Path(__file__).parent / "blogdb/sql"


@pytest.fixture
def sql_file(sql_dir):
    return sql_dir / "blogs/blogs.sql"


@pytest.fixture
def sql(sql_file):
    with open(sql_file) as f:
        return f.read()


def test_version():
    assert re.match(r"\d+\.\d+(\.?dev\d*)?$", aiosql.__version__)


def test_frompath_queries_cls(sql_dir):
    class TestQueries(Queries):
        pass

    queries = aiosql.from_path(sql_dir, "aiosqlite", queries_cls=TestQueries)
    assert isinstance(queries, TestQueries)

    assert repr(queries).startswith("Queries(")


def test_frompath_queryloader_cls(sql_dir):
    mock_loader = mock.MagicMock(wraps=QueryLoader)

    aiosql.from_path(sql_dir, "aiosqlite", loader_cls=mock_loader)

    assert mock_loader.called


def test_fromstr_queries_cls(sql):
    class TestQueries(Queries):
        pass

    queries = aiosql.from_str(sql, "aiosqlite", queries_cls=TestQueries)
    assert isinstance(queries, TestQueries)


def test_fromstr_queryloader_cls(sql):
    mock_loader = mock.MagicMock(wraps=QueryLoader)

    aiosql.from_str(sql, "aiosqlite", loader_cls=mock_loader)

    assert mock_loader.called


def test_trailing_space_on_lines_does_not_error():
    # There is whitespace in this string after the line ends
    sql_str = "-- name: trailing-space^    \n"
    sql_str += "select * from test;     \n"

    try:
        aiosql.from_str(sql_str, "aiosqlite")
    except SQLParseException:  # pragma: no cover
        pytest.fail("Raised SQLParseException due to trailing space in query.")


def test_non_ascii_char():
    # this triggers a warning, that we do not really check but for coverage
    q = aiosql.from_str("-- name: zéro\nSELECT 0;\n", "sqlite3")
    assert "zéro" in q._available_queries
    q = aiosql.from_str("-- name: éêèëÉÊÈË\nSELECT 'eeeeeeee!';\n", "sqlite3")
    assert "éêèëÉÊÈË" in q._available_queries
    q = aiosql.from_str("-- name: 안녕하세요\nSELECT 'hello!';\n", "sqlite3")
    assert "안녕하세요" in q._available_queries


def test_loading_query_signature():
    sql_str = "-- name: get^\n" "select * from test where foo=:foo and bar=:bar"
    queries = aiosql.from_str(sql_str, "aiosqlite")
    assert queries.get.__signature__ == inspect.Signature(
        [
            inspect.Parameter("self", kind=inspect.Parameter.POSITIONAL_OR_KEYWORD),
            inspect.Parameter("foo", kind=inspect.Parameter.KEYWORD_ONLY),
            inspect.Parameter("bar", kind=inspect.Parameter.KEYWORD_ONLY),
        ]
    )


def test_names():
    try:
        queries = aiosql.from_str("-- name: 1st\nSELECT 1;\n", "sqlite3")
        assert False, "'1st' should be rejected"
    except SQLParseException as e:
        assert '"1st"' in str(e)
    try:
        queries = aiosql.from_str("-- name: one$garbage\nSELECT 1;\n", "sqlite3")
        assert False, "garbage should be rejected"
    except SQLParseException as e:
        assert 'garbage"' in str(e)
    # - is okay because mapped to _
    queries = aiosql.from_str("-- name: -dash\nSELECT 1;\n", "sqlite3")
    assert "_dash" in queries.available_queries


def test_loading_query_signature_with_duplicate_parameter():
    sql_str = "-- name: get^\n" "select * from test where foo=:foo and foo=:foo"
    queries = aiosql.from_str(sql_str, "aiosqlite")
    assert queries.get.__signature__ == inspect.Signature(
        [
            inspect.Parameter("self", kind=inspect.Parameter.POSITIONAL_OR_KEYWORD),
            inspect.Parameter("foo", kind=inspect.Parameter.KEYWORD_ONLY),
        ]
    )


def test_adapters():
    try:
        aiosql.aiosql._make_driver_adapter("no-such-driver-adapter")
        assert False, "must raise an exception"  # pragma: no cover
    except ValueError as e:
        assert "unregistered driver_adapter" in str(e)

    class PyFormatConnector:
        paramstyle = "pyformat"

    a = aiosql.aiosql._make_driver_adapter(PyFormatConnector)
    assert type(a) == aiosql.adapters.PyFormatAdapter

    class NamedConnector:
        paramstyle = "named"

    a = aiosql.aiosql._make_driver_adapter(NamedConnector)
    assert type(a) == aiosql.adapters.GenericAdapter

    class NoSuchConnector:
        paramstyle = "no-such-style"

    try:
        aiosql.aiosql._make_driver_adapter(NoSuchConnector)
        assert False, "must raise an exception"  # pragma: no cover
    except ValueError as e:
        assert "Unexpected driver_adapter" in str(e)

    try:
        aiosql.aiosql._make_driver_adapter(True)  # type: ignore
        assert False, "must raise an exception"  # pragma: no cover
    except ValueError as e:
        assert "Unexpected driver_adapter" in str(e)


def test_no_such_path():
    try:
        aiosql.from_path("/no/such/file", "sqlite3")
        assert False, "must raise an exception"  # pragma: no cover
    except SQLLoadException as e:
        assert "File does not exist" in str(e)


def test_file_loading(sql_file):
    db = aiosql.from_path(sql_file, "sqlite3")
    assert "get_user_blogs" in db.__dict__


def test_misc(sql_file):
    try:
        queries = aiosql.queries.Queries("sqlite3")
        queries._make_sync_fn(("hello", None, -1, "SELECT NULL;", None, None, None))
        assert False, "must raise an exception"  # pragma: no cover
    except ValueError as e:
        assert "Unknown operation_type" in str(e)
    try:
        db = aiosql.from_str("-- name: a*b\nSELECT 'ab'\n", "sqlite3")
        assert False, "must raise en exception"  # pragma: no cover
    except Exception as e:
        assert "invalid query name and operation" in str(e)
    ql = aiosql.query_loader.QueryLoader(None, None)
    try:
        ql.load_query_data_from_dir_path(sql_file)
        assert False, "must raise en exception"  # pragma: no cover
    except ValueError as e:
        assert "must be a directory" in str(e)


def test_kwargs():
    # kwargs_only == True
    queries = aiosql.from_str("-- name: plus_one$\nSELECT 1 + :val;\n", "sqlite3", kwargs_only=True)
    import sqlite3

    conn = sqlite3.connect(":memory:")
    assert 42 == queries.plus_one(conn, val=41)
    try:
        queries.plus_one(conn, 2)
        assert False, "must raise an exception"  # pragma: no cover
    except ValueError as e:
        assert "kwargs" in str(e)
    # kwargs_only == False
    queries = aiosql.from_str(
        "-- name: plus_two$\nSELECT 2 + :val;\n", "sqlite3", kwargs_only=False
    )
    assert 42 == queries.plus_two(conn, val=40)
    try:
        queries.plus_two(conn, 2, val=41)
        assert False, "must raise an exception"  # pragma: no cover
    except ValueError as e:
        assert "mix" in str(e)



import pytest
import sqlite3
from pathlib import Path
from conf_schema import create_user_blogs, fill_user_blogs


def populate_sqlite3_db(db_path):
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    cur.executescript("\n".join(create_user_blogs("sqlite")))
    fill_user_blogs(cur, "sqlite")
    conn.commit()
    conn.close()


@pytest.fixture
def sqlite3_db_path(tmpdir):
    db_path = str(Path(tmpdir.strpath) / "blogdb.db")
    populate_sqlite3_db(db_path)
    return db_path



import pytest
import aiosql

from conf_mysql import my_dsn, my_db, my_conn
from conf_pgsql import pg_conn, pg_params, pg_dsn
from conf_sqlite import sqlite3_db_path
from conf_duckdb import duckdb_db_path, duckdb_conn


def pytest_addoption(parser):
    parser.addoption("--postgresql-detached", action="store_true")
    parser.addoption("--mysql-detached", action="store_true")
    parser.addoption("--mysql-tries", default=1, type=int)
    parser.addoption(
        "--mysql-driver",
        default="MySQLdb",
        choices=["MySQLdb", "mysql.connector", "pymysql", "mariadb"],
        help="which driver to use for creating connections",
    )


# test adapter registering and overriding
aiosql.aiosql.register_adapter("named", aiosql.adapters.GenericAdapter)
aiosql.register_adapter("named", aiosql.adapters.GenericAdapter)



import aiosql
import pytest
import run_tests as t
import sqlite3 as db

pytestmark = [pytest.mark.sqlite3]

DRIVER = "sqlite3"


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


@pytest.fixture
def queries():
    return t.queries(DRIVER)


@pytest.fixture
def conn(sqlite3_db_path):
    conn = db.connect(sqlite3_db_path)
    yield conn
    conn.close()


def test_cursor(conn, queries):
    t.run_cursor(conn, queries)


def test_record_query(conn, queries):
    conn.row_factory = dict_factory
    t.run_record_query(conn, queries)


def test_parameterized_query(conn, queries):
    t.run_parameterized_query(conn, queries)


def test_parameterized_record_query(conn, queries):
    conn.row_factory = dict_factory
    t.run_parameterized_record_query(conn, queries, DRIVER, t.todate)


def test_record_class_query(conn, queries):
    t.run_record_class_query(conn, queries, t.todate)


def test_select_cursor_context_manager(conn, queries):
    t.run_select_cursor_context_manager(conn, queries, t.todate)


def test_select_one(conn, queries):
    t.run_select_one(conn, queries)


def test_select_value(conn, queries):
    t.run_select_value(conn, queries, DRIVER)


def test_modulo(conn, queries):
    actual = queries.blogs.sqlite_get_modulo(conn, numerator=7, denominator=3)
    expected = 7 % 3
    assert actual == expected


def test_insert_returning(conn, queries):
    t.run_insert_returning(conn, queries, DRIVER, t.todate)


def test_delete(conn, queries):
    t.run_delete(conn, queries)


def test_insert_many(conn, queries):
    with conn:
        t.run_insert_many(conn, queries, t.todate)


def test_date_time(conn, queries):
    t.run_date_time(conn, queries, DRIVER)


def test_execute_script(conn, queries):
    with conn:
        actual = queries.comments.sqlite_create_comments_table(conn)
        assert actual == "DONE"



from .generic import GenericAdapter
from ..utils import VAR_REF


def _replacer(match):
    """Regex hook for named to pyformat conversion."""
    gd = match.groupdict()
    if gd["dquote"] is not None:  # "..."
        return gd["dquote"]
    elif gd["squote"] is not None:  # '...'
        return gd["squote"]
    else:  # :something to %(something)s
        return f'{gd["lead"]}%({gd["var_name"]})s'


class PyFormatAdapter(GenericAdapter):
    """Convert from named to pyformat parameter style."""

    def process_sql(self, _query_name, _op_type, sql):
        """From named to pyformat."""
        return VAR_REF.sub(_replacer, sql)



import csv
from pathlib import Path

# CSV data file paths
BLOGDB_PATH = Path(__file__).parent / "blogdb"
USERS_DATA_PATH = BLOGDB_PATH / "data/users_data.csv"
BLOGS_DATA_PATH = BLOGDB_PATH / "data/blogs_data.csv"


def create_user_blogs(db):
    assert db in ("sqlite", "pgsql", "mysql", "duckdb")
    serial = (
        "SERIAL"
        if db == "pgsql"
        else "INTEGER" if db in ("sqlite", "duckdb") else "INTEGER auto_increment"
    )
    ddl_statements = [
        f"""CREATE TABLE IF NOT EXISTS users (
                userid {serial} PRIMARY KEY,
                username TEXT NOT NULL,
                firstname TEXT NOT NULL,
                lastname TEXT NOT NULL);""",
        f"""CREATE TABLE IF NOT EXISTS blogs (
                blogid {serial} PRIMARY KEY,
                userid INTEGER NOT NULL,
                title TEXT NOT NULL,
                content TEXT NOT NULL,
                published DATE NOT NULL DEFAULT (CURRENT_DATE),
                FOREIGN KEY (userid) REFERENCES users(userid));""",
    ]
    if db and db == "duckdb":
        ddl_statements.extend(["create sequence users_seq;", "create sequence blogs_seq;"])
    return tuple(ddl_statements)


def drop_user_blogs(db):
    return (
        "DROP TABLE IF EXISTS comments",
        "DROP TABLE IF EXISTS blogs",
        "DROP TABLE IF EXISTS users",
    )


def fill_user_blogs(cur, db):
    # NOTE postgres filling relies on copy
    assert db in ("sqlite", "mysql")
    param = "?" if db == "sqlite" else "%s"
    with USERS_DATA_PATH.open() as fp:
        users = list(csv.reader(fp))
        cur.executemany(
            f"""
               INSERT INTO users (
                    username,
                    firstname,
                    lastname
               ) VALUES ({param}, {param}, {param});""",
            users,
        )
    with BLOGS_DATA_PATH.open() as fp:
        blogs = list(csv.reader(fp))
        cur.executemany(
            f"""
                INSERT INTO blogs (
                    userid,
                    title,
                    content,
                    published
                ) VALUES ({param}, {param}, {param}, {param});""",
            blogs,
        )



import aiosql
import pytest
from aiosql.queries import Queries
import run_tests as t
import utils as u
from datetime import date

try:
    import duckdb as db
except ModuleNotFoundError:
    pytest.skip("missing driver: duckdb", allow_module_level=True)

pytestmark = [pytest.mark.duckdb]

DRIVER = "duckdb"


@pytest.fixture
def conn(duckdb_conn):
    return duckdb_conn


@pytest.fixture
def queries() -> Queries:
    return t.queries(DRIVER)


def test_cursor(conn, queries):
    t.run_cursor(conn, queries)


@pytest.mark.skip("does not work yet")
def test_record_query(conn, queries: Queries):
    queries.driver_adapter.convert_row_to_dict = True
    t.run_record_query(conn, queries)


def test_parameterized_query(conn, queries):
    t.run_parameterized_query(conn, queries, DRIVER)


@pytest.mark.skip("does not work yet")
def test_parameterized_record_query(conn, queries):
    # queries.driver_adapter.convert_row_to_dict = True
    t.run_parameterized_record_query(conn, queries, DRIVER, date)


@pytest.mark.skip("does not work yet")
def test_record_class_query(conn, queries):
    t.run_record_class_query(conn, queries, t.todate, DRIVER)


def test_select_cursor_context_manager(conn, queries):
    t.run_select_cursor_context_manager(conn, queries, date, DRIVER)


def test_select_one(conn, queries):
    t.run_select_one(conn, queries, DRIVER)


def test_select_value(conn, queries):
    t.run_select_value(conn, queries, db=DRIVER)


def test_modulo(conn, queries):
    actual = queries.blogs.sqlite_get_modulo(conn, numerator=7, denominator=3)
    expected = 7 % 3
    assert actual == expected


@pytest.mark.skip("does not work anymore, on version 0.9.0")
def test_insert_returning(conn, queries):
    t.run_insert_returning(conn, queries, db=DRIVER, todate=t.todate)


def test_delete(conn, queries):
    t.run_delete(conn, queries, db=DRIVER)


@pytest.mark.skip("does not work yet: execute many does not support named parameters?")
def test_insert_many(conn, queries):
    with conn:
        t.run_insert_many(conn, queries, t.todate, db=DRIVER)


def test_date_time(conn, queries):
    t.run_date_time(conn, queries, db=DRIVER)


def test_execute_script(conn, queries):
    with conn:
        actual = queries.comments.duckdb_create_comments_table(conn)
        assert actual == "DONE"



from contextlib import contextmanager
from typing import List


class GenericAdapter:
    """
    Generic AioSQL Adapter suitable for `named` parameter style and no with support.

    This class also serves as the base class for other adapters.
    """

    def __init__(self, driver=None):
        self._driver = driver

    def process_sql(self, _query_name, _op_type, sql):
        """Preprocess SQL query."""
        return sql

    def _cursor(self, conn):
        """Get a cursor from a connection."""
        return conn.cursor()

    def select(self, conn, _query_name: str, sql: str, parameters, record_class=None):
        """Handle a relation-returning SELECT (no suffix)."""
        column_names: List[str] = []
        cur = self._cursor(conn)
        try:
            cur.execute(sql, parameters)
            if record_class is None:
                for row in cur:
                    yield row
            else:
                first = True
                for row in cur:
                    if first:  # only get description on the fly, for apsw
                        column_names = [c[0] for c in cur.description]
                        first = False
                    yield record_class(**dict(zip(column_names, row)))
        finally:
            cur.close()

    def select_one(self, conn, _query_name, sql, parameters, record_class=None):
        """Handle a tuple-returning (one row) SELECT (``^`` suffix).

        Return None if empty."""
        cur = self._cursor(conn)
        try:
            cur.execute(sql, parameters)
            result = cur.fetchone()
            if result is not None and record_class is not None:
                column_names = [c[0] for c in cur.description]
                # this fails if result is not a list or tuple
                result = record_class(**dict(zip(column_names, result)))
        finally:
            cur.close()
        return result

    def select_value(self, conn, _query_name, sql, parameters):
        """Handle a scalar-returning (one value) SELECT (``$`` suffix).

        Return None if empty."""
        cur = self._cursor(conn)
        try:
            cur.execute(sql, parameters)
            result = cur.fetchone()
            if result:
                if isinstance(result, (list, tuple)):
                    return result[0]
                elif isinstance(result, dict):
                    return next(iter(result.values()))
                else:  # pragma: no cover
                    raise Exception(f"unexpected value type: {type(result)}")
            else:
                return None
        finally:
            cur.close()

    @contextmanager
    def select_cursor(self, conn, _query_name, sql, parameters):
        """Return the raw cursor after a SELECT exec."""
        cur = self._cursor(conn)
        cur.execute(sql, parameters)
        try:
            yield cur
        finally:
            cur.close()

    def insert_update_delete(self, conn, _query_name, sql, parameters):
        """Handle affected row counts (INSERT UPDATE DELETE) (``!`` suffix)."""
        cur = self._cursor(conn)
        cur.execute(sql, parameters)
        rc = cur.rowcount if hasattr(cur, "rowcount") else -1
        cur.close()
        return rc

    def insert_update_delete_many(self, conn, _query_name, sql, parameters):
        """Handle affected row counts (INSERT UPDATE DELETE) (``*!`` suffix)."""
        cur = self._cursor(conn)
        cur.executemany(sql, parameters)
        rc = cur.rowcount if hasattr(cur, "rowcount") else -1
        cur.close()
        return rc

    # FIXME this made sense when SQLite had no RETURNING prefix (v3.35, 2021-03-12)
    def insert_returning(self, conn, _query_name, sql, parameters):
        """Special case for RETURNING (``<!`` suffix) with SQLite."""
        # very similar to select_one but the returned value
        cur = self._cursor(conn)
        cur.execute(sql, parameters)
        res = cur.fetchone()
        cur.close()
        return res[0] if res and len(res) == 1 else res

    def execute_script(self, conn, sql):
        """Handle an SQL script (``#`` suffix)."""
        cur = self._cursor(conn)
        cur.execute(sql)
        msg = cur.statusmessage if hasattr(cur, "statusmessage") else "DONE"
        cur.close()
        return msg



from datetime import date

import pytest
import aiosql
import run_tests as t
import utils as u

try:
    import asyncpg
except ModuleNotFoundError:
    pytest.skip("missing driver: asyncpg", allow_module_level=True)

pytestmark = [
    pytest.mark.postgres,
    # pytest.mark.asyncio,
    pytest.mark.skipif(not u.has_pkg("pytest_postgresql"), reason="no pytest_postgresql"),
    pytest.mark.skipif(not u.has_pkg("pytest_asyncio"), reason="no pytest_asyncio"),
]

DRIVER = "asyncpg"


@pytest.fixture
def queries():
    return t.queries(DRIVER)


@pytest.mark.asyncio
async def test_record_query(pg_dsn, queries):
    conn = await asyncpg.connect(pg_dsn)
    await t.run_async_record_query(conn, queries)
    await conn.close()


@pytest.mark.asyncio
async def test_parameterized_query(pg_dsn, queries):
    conn = await asyncpg.connect(pg_dsn)
    actual = await t.run_async_parameterized_query(conn, queries, date)
    await conn.close()


@pytest.mark.asyncio
async def test_many_replacements(pg_dsn, queries):
    """If the replacement was longer than the variable, bad SQL was generated.

    The variable replacement code had a bug that caused it to miscalculate where in the
    original string to put the placeholders.  The SQL below would produce a query that
    ended with "$8, $9, $10$11:k);" because of this bug.

    This test would fail before the bug was fixed and passes afterward.

    This issue was reported in https://github.com/nackjicholson/aiosql/issues/90.
    """
    sql = r"""
        -- name: test<!
        INSERT INTO table VALUES (:a, :b, :c, :d, :e, :f, :g, :h, :i, :j, :k);
    """
    actual = aiosql.from_str(sql, "asyncpg").test.sql
    expected = "INSERT INTO table " "VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);"
    assert actual == expected


@pytest.mark.asyncio
async def test_parameterized_record_query(pg_dsn, queries):
    conn = await asyncpg.connect(pg_dsn)
    await t.run_async_parameterized_record_query(conn, queries, DRIVER, date)
    await conn.close()


@pytest.mark.asyncio
async def test_record_class_query(pg_dsn, queries):
    conn = await asyncpg.connect(pg_dsn)
    await t.run_async_record_class_query(conn, queries, date)
    await conn.close()


@pytest.mark.asyncio
async def test_select_cursor_context_manager(pg_dsn, queries):
    conn = await asyncpg.connect(pg_dsn)
    await t.run_async_select_cursor_context_manager(conn, queries, date)
    await conn.close()


@pytest.mark.asyncio
async def test_select_one(pg_dsn, queries):
    conn = await asyncpg.connect(pg_dsn)
    await t.run_async_select_one(conn, queries)
    await conn.close()


@pytest.mark.asyncio
async def test_select_value(pg_dsn, queries):
    conn = await asyncpg.connect(pg_dsn)
    await t.run_async_select_value(conn, queries)
    await conn.close()


@pytest.mark.asyncio
async def test_insert_returning(pg_dsn, queries):
    async with asyncpg.create_pool(pg_dsn) as pool:
        async with pool.acquire() as conn:
            await t.run_async_insert_returning(conn, queries, DRIVER, date)

    conn = await asyncpg.connect(pg_dsn)
    res = await queries.blogs.pg_no_publish(conn)
    assert res is None
    await conn.close()


@pytest.mark.asyncio
async def test_delete(pg_dsn, queries):
    conn = await asyncpg.connect(pg_dsn)
    await t.run_async_delete(conn, queries)
    await conn.close()


@pytest.mark.asyncio
async def test_insert_many(pg_dsn, queries):
    conn = await asyncpg.connect(pg_dsn)
    await t.run_async_insert_many(conn, queries, date)
    await conn.close()


@pytest.mark.asyncio
async def test_async_methods(pg_dsn, queries):
    async with asyncpg.create_pool(pg_dsn) as pool:
        await t.run_async_methods(pool, queries)


@pytest.mark.asyncio
async def test_execute_script(pg_dsn, queries):
    async with asyncpg.create_pool(pg_dsn) as pool:
        actual = await queries.comments.pg_create_comments_table(pool, queries)
        assert actual == "CREATE TABLE"


@pytest.mark.asyncio
async def test_variable_replacement(pg_dsn, queries):
    # Addresses bug reported in the following github issue:
    # https://github.com/nackjicholson/aiosql/issues/51
    #
    # When the test fails the users_sql below will fail to increment the $2
    # variable for lastname.
    async with asyncpg.create_pool(pg_dsn) as pool:
        blogs_sql = "select title from blogs where title = $1 and published = $2;"
        assert queries.blogs.search.sql == blogs_sql
        users_sql = "select username from users where firstname = $1 and lastname = $2;"
        assert queries.users.search.sql == users_sql

        blogs_res = await queries.blogs.search(pool, title="Testing", published=date(2018, 1, 1))
        users_res = await queries.users.search(pool, title="John", lastname="Doe")
        assert blogs_res == [("Testing",)]
        assert users_res == [("johndoe",)]

        square_sql = "select $1::int * $1::int as squared;"
        assert queries.blogs.square.sql == square_sql

        square_res = await queries.blogs.square(pool, val=42)
        assert square_res == 1764


def test_maybe_order_params():
    a = aiosql.adapters.asyncpg.AsyncPGAdapter()
    try:
        a.maybe_order_params("foo", "wrong-type-parameter")
        assert False, "exception should be raised"  # pragma: no cover
    except ValueError as e:
        assert "dict or tuple" in str(e)



import re
import inspect
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Type, Sequence, Any, Union

from .utils import SQLParseException, SQLLoadException, VAR_REF, log
from .types import QueryDatum, QueryDataTree, SQLOperationType, DriverAdapterProtocol

# identifies name definition comments
_QUERY_DEF = re.compile(r"--\s*name\s*:\s*")

# identifies record class definition comments
_RECORD_DEF = re.compile(r"--\s*record_class\s*:\s*(\w+)\s*")

# extract a valid query name followed by an optional operation spec
# FIXME this accepts "1st" but seems to reject "é"
_NAME_OP = re.compile(r"^(\w+)(|\^|\$|!|<!|\*!|#)$")

# forbid numbers as first character
_BAD_PREFIX = re.compile(r"^\d")

# get SQL comment contents
_SQL_COMMENT = re.compile(r"\s*--\s*(.*)$")

# map operation suffixes to their type
_OP_TYPES = {
    "<!": SQLOperationType.INSERT_RETURNING,
    "*!": SQLOperationType.INSERT_UPDATE_DELETE_MANY,
    "!": SQLOperationType.INSERT_UPDATE_DELETE,
    "#": SQLOperationType.SCRIPT,
    "^": SQLOperationType.SELECT_ONE,
    "$": SQLOperationType.SELECT_VALUE,
    "": SQLOperationType.SELECT,
}

# extracting comments requires some kind of scanner
_UNCOMMENT = re.compile(
    # single quote strings
    r"(?P<squote>\'(\'\'|[^\'])*\')|"
    # double quote strings
    r'(?P<dquote>"(""|[^"])+")|'
    # one-line comment
    r"(?P<oneline>--.*?$)|"
    # multi-line comment
    r"(?P<multiline>/\*.*?\*/)",
    re.DOTALL | re.MULTILINE,
)


def _remove_ml_comments(code: str) -> str:
    """Remove /* ... */ comments from code"""
    # identify commented regions
    rm = []
    for m in _UNCOMMENT.finditer(code):
        ml = m.groupdict()["multiline"]
        if ml:
            rm.append(m.span())
    # keep whatever else
    ncode, current = "", 0
    for start, end in rm:
        ncode += code[current:start]
        current = end
    ncode += code[current:]
    return ncode


class QueryLoader:
    def __init__(
        self, driver_adapter: DriverAdapterProtocol, record_classes: Optional[Dict[str, Any]]
    ):
        self.driver_adapter = driver_adapter
        self.record_classes = record_classes if record_classes is not None else {}

    def _make_query_datum(
        self, query: str, ns_parts: List[str], floc: Tuple[Union[Path, str], int]
    ) -> QueryDatum:
        # Build a query datum
        # - query: the spec and name ("query-name!\n-- comments\nSQL;\n")
        # - ns_parts: name space parts, i.e. subdirectories of loaded files
        # - floc: file name and lineno the query was extracted from
        lines = [line.strip() for line in query.strip().splitlines()]
        qname, qop = self._get_name_op(lines[0])
        if re.search(r"[^A-Za-z0-9_]", qname):
            log.warning(f"non ASCII character in query name: {qname}")
        record_class = self._get_record_class(lines[1])
        sql, doc = self._get_sql_doc(lines[2 if record_class else 1 :])
        signature = self._build_signature(sql)
        query_fqn = ".".join(ns_parts + [qname])
        sql = self.driver_adapter.process_sql(query_fqn, qop, sql)
        return QueryDatum(query_fqn, doc, qop, sql, record_class, signature, floc)

    def _get_name_op(self, text: str) -> Tuple[str, SQLOperationType]:
        qname_spec = text.replace("-", "_")
        nameop = _NAME_OP.match(qname_spec)
        if not nameop or _BAD_PREFIX.match(qname_spec):
            raise SQLParseException(f'invalid query name and operation spec: "{qname_spec}"')
        qname, qop = nameop.group(1, 2)
        return qname, _OP_TYPES[qop]

    def _get_record_class(self, text: str) -> Optional[Type]:
        rc_match = _RECORD_DEF.match(text)
        rc_name = rc_match.group(1) if rc_match else None
        # TODO: Probably will want this to be a class, marshal in, and marshal out
        return self.record_classes.get(rc_name) if isinstance(rc_name, str) else None

    def _get_sql_doc(self, lines: Sequence[str]) -> Tuple[str, str]:
        doc, sql = "", ""
        for line in lines:
            doc_match = _SQL_COMMENT.match(line)
            if doc_match:
                doc += doc_match.group(1) + "\n"
            else:
                sql += line + "\n"

        return sql.strip(), doc.rstrip()

    def _build_signature(self, sql: str) -> inspect.Signature:
        params = [inspect.Parameter("self", inspect.Parameter.POSITIONAL_OR_KEYWORD)]
        names = set()
        for match in VAR_REF.finditer(sql):
            gd = match.groupdict()
            if gd["squote"] or gd["dquote"]:
                continue
            name = gd["var_name"]
            if name.isdigit() or name in names:
                continue
            names.add(name)
            params.append(
                inspect.Parameter(
                    name=name,
                    kind=inspect.Parameter.KEYWORD_ONLY,
                )
            )
        return inspect.Signature(parameters=params)

    def load_query_data_from_sql(
        self, sql: str, ns_parts: List[str], fname: Union[Path, str] = "<unknown>"
    ) -> List[QueryDatum]:
        usql = _remove_ml_comments(sql)
        qdefs = _QUERY_DEF.split(usql)
        # FIXME lineno is from the uncommented file
        lineno = 1 + qdefs[0].count("\n")
        data = []
        # first item is anything before the first query definition, drop it!
        for qdef in qdefs[1:]:
            data.append(self._make_query_datum(qdef, ns_parts, (fname, lineno)))
            lineno += qdef.count("\n")
        return data

    def load_query_data_from_file(
        self, path: Path, ns_parts: List[str] = [], encoding=None
    ) -> List[QueryDatum]:
        return self.load_query_data_from_sql(path.read_text(encoding=encoding), ns_parts, path)

    def load_query_data_from_dir_path(
        self, dir_path, ext=(".sql",), encoding=None
    ) -> QueryDataTree:
        if not dir_path.is_dir():
            raise ValueError(f"The path {dir_path} must be a directory")

        def _recurse_load_query_data_tree(path, ns_parts=[], ext=(".sql",), encoding=None):
            query_data_tree = {}
            for p in path.iterdir():
                if p.is_file():
                    if p.suffix not in ext:
                        continue
                    for query_datum in self.load_query_data_from_file(
                        p, ns_parts, encoding=encoding
                    ):
                        query_data_tree[query_datum.query_name] = query_datum
                elif p.is_dir():
                    query_data_tree[p.name] = _recurse_load_query_data_tree(
                        p, ns_parts + [p.name], ext=ext, encoding=encoding
                    )
                else:  # pragma: no cover
                    # This should be practically unreachable.
                    raise SQLLoadException(f"The path must be a directory or file, got {p}")
            return query_data_tree

        return _recurse_load_query_data_tree(dir_path, ext=ext, encoding=encoding)



from pathlib import Path
from typing import NamedTuple, Iterable
from datetime import date
import asyncio
import re

import aiosql
from utils import log


# for sqlite3
def todate(year, month, day):
    return f"{year:04}-{month:02}-{day:02}"


class UserBlogSummary(NamedTuple):
    title: str
    published: date


# map drivers to databases
_DB = {
    "sqlite3": "sqlite3",
    "apsw": "sqlite3",
    "aiosqlite": "sqlite3",
    "psycopg": "postgres",
    "psycopg2": "postgres",
    "asyncpg": "postgres",
    "pygresql": "postgres",
    "pg8000": "postgres",
    "pymysql": "mysql",
    "mysql-connector": "mysql",
    "mysqldb": "mysql",
    "mariadb": "mariadb",
    "duckdb": "duckdb",
}


RECORD_CLASSES = {"UserBlogSummary": UserBlogSummary}


def queries(driver):
    """Load queries into AioSQL."""
    dir_path = Path(__file__).parent / "blogdb" / "sql"
    return aiosql.from_path(dir_path, driver, RECORD_CLASSES)


def run_something(conn):
    """Run something on a connection without a schema."""

    def sel12(cur):
        cur.execute("SELECT 1, 'un' UNION SELECT 2, 'deux' ORDER BY 1")
        res = cur.fetchall()
        assert type(res) in (tuple, list), f"unexpected type: {type(res)}"
        assert res == ((1, "un"), (2, "deux")) or res == [(1, "un"), (2, "deux")]

    cur = conn.cursor()
    has_with = hasattr(cur, "__enter__")
    sel12(cur)
    cur.close()

    if has_with:  # if available
        with conn.cursor() as cur:
            sel12(cur)

    conn.commit()


def run_cursor(conn, queries):
    cur = queries.driver_adapter._cursor(conn)
    cur.execute("SELECT 'Hello World!'")
    res = cur.fetchone()
    assert res in [("Hello World!",), ["Hello World!"]]
    cur.close()


def run_record_query(conn, queries):
    raw_actual = queries.users.get_all(conn)
    assert isinstance(raw_actual, Iterable)
    actual = list(raw_actual)
    assert len(actual) == 3
    assert actual[0] == {
        "userid": 1,
        "username": "bobsmith",
        "firstname": "Bob",
        "lastname": "Smith",
    }


def run_parameterized_query(conn, queries, db=None):
    # select on a parameter
    actual = queries.users.get_by_lastname(conn, lastname="Doe")
    expected = [(3, "janedoe", "Jane", "Doe"), (2, "johndoe", "John", "Doe")]
    # NOTE re-conversion needed for mysqldb and pg8000
    actual = [tuple(i) for i in actual]
    assert actual == expected

    # select with 3 parameters
    # FIXME broken with pg8000
    # actual = queries.misc.comma_nospace_var(conn, one=1, two=10, three=100)
    # assert actual == (1, 10, 100) or actual == [1, 10, 100]
    actual = queries.misc.comma_nospace_var(conn, one="Hello", two=" ", three="World!")

    # NOTE some drivers return a list instead of a tuple
    assert actual == ("Hello", " ", "World!") or actual == ["Hello", " ", "World!"]


def run_parameterized_record_query(conn, queries, db, todate):
    if _DB[db] == "sqlite3":
        fun = queries.blogs.sqlite_get_blogs_published_after
    elif _DB[db] == "duckdb":
        fun = queries.blogs.duckdb_get_blogs_published_after
    elif _DB[db] == "postgres":
        fun = queries.blogs.pg_get_blogs_published_after
    elif _DB[db] in ("mysql", "mariadb"):
        fun = queries.blogs.my_get_blogs_published_after
    else:
        raise Exception(f"unexpected driver: {db}")

    raw_actual = fun(conn, published=todate(2018, 1, 1))
    assert isinstance(raw_actual, Iterable)
    actual = list(raw_actual)

    expected = [
        {
            "title": "How to make a pie.",
            "username": "bobsmith",
            "published": "2018-11-23 00:00",
        },
        {"title": "Testing", "username": "janedoe", "published": "2018-01-01 00:00"},
    ]

    assert actual == expected


def run_record_class_query(conn, queries, todate, db=None):
    raw_actual = queries.blogs.get_user_blogs(conn, userid=1)
    assert isinstance(raw_actual, Iterable)
    actual = list(raw_actual)

    expected = [
        UserBlogSummary(title="How to make a pie.", published=todate(2018, 11, 23)),
        UserBlogSummary(title="What I did Today", published=todate(2017, 7, 28)),
    ]
    assert all(isinstance(row, UserBlogSummary) for row in actual)
    assert actual == expected

    one = queries.blogs.get_latest_user_blog(conn, userid=1)
    assert one == UserBlogSummary(title="How to make a pie.", published=todate(2018, 11, 23))


def run_select_cursor_context_manager(conn, queries, todate, db=None):
    fun = queries.blogs.get_user_blogs_cursor
    expected = [
        ("How to make a pie.", todate(2018, 11, 23)),
        ("What I did Today", todate(2017, 7, 28)),
    ]

    with fun(conn, userid=1) as cursor:
        # reconversions for mysqldb and pg8000
        actual = [tuple(r) for r in cursor.fetchall()]
        assert actual == expected


def run_select_one(conn, queries, db=None):
    actual = queries.users.get_by_username(conn, username="johndoe")

    # reconversion for pg8000
    actual = tuple(actual)
    expected = (2, "johndoe", "John", "Doe")
    assert actual == expected


def run_insert_returning(conn, queries, db, todate):
    if _DB[db] in ("sqlite3"):
        fun = queries.blogs.publish_blog
    elif _DB[db] in ("duckdb"):
        fun = queries.blogs.duckdb_publish_blog
    elif _DB[db] in ("postgres", "mariadb"):
        fun = queries.blogs.pg_publish_blog
    elif _DB[db] == "mysql":
        fun = queries.blogs.my_publish_blog
    else:
        raise Exception(f"unexpected driver: {db}")

    if db == "duckdb":
        blogid = fun(
            conn,
            2,
            "My first blog",
            "Hello, World!",
            todate(2018, 12, 4),
        )
    else:
        blogid = fun(
            conn,
            userid=2,
            title="My first blog",
            contents="Hello, World!",
            published=todate(2018, 12, 4),
        )

    # sqlite returns a number while pg query returns a tuple
    if isinstance(blogid, tuple):
        assert db in ("psycopg", "psycopg2", "pygresql", "mariadb", "duckdb")
        blogid, title = blogid
    elif isinstance(blogid, list):
        assert db == "pg8000"
        blogid, title = blogid
    # duckdb will return a tuple or a dict depending on how the sql is structured.
    # If you wrap the returning in `()` a dict is returned.
    elif isinstance(blogid, dict):
        assert db == "duckdb"
        title = blogid.get("title")
        blogid = blogid.get("blogid")
    else:
        assert db in ("sqlite3", "apsw")
        blogid, title = blogid, "My first blog"

    b2, t2 = queries.blogs.blog_title(conn, blogid=blogid)
    assert (blogid, title) == (b2, t2)

    if db and db in ("psycopg", "psycopg2"):
        res = queries.blogs.pg_no_publish(conn)
        assert res is None


def run_delete(conn, queries, expect=1, db=None):
    # Removing the "janedoe" blog titled "Testing"
    actual = queries.blogs.remove_blog(conn, blogid=2)
    raw_janes_blogs = queries.blogs.get_user_blogs(conn, userid=3)
    assert actual in (expect, -1)
    assert isinstance(raw_janes_blogs, Iterable)

    janes_blogs = list(raw_janes_blogs)
    assert len(janes_blogs) == 0


def run_insert_many(conn, queries, todate, expect=3, db=None):
    blogs = [
        {
            "userid": 2,
            "title": "Blog Part 1",
            "contents": "content - 1",
            "published": todate(2018, 12, 4),
        },
        {
            "userid": 2,
            "title": "Blog Part 2",
            "contents": "content - 2",
            "published": todate(2018, 12, 5),
        },
        {
            "userid": 2,
            "title": "Blog Part 3",
            "contents": "content - 3",
            "published": todate(2018, 12, 6),
        },
    ]

    actual = queries.blogs.pg_bulk_publish(conn, blogs)
    raw_johns_blogs = queries.blogs.get_user_blogs(conn, userid=2)
    expected = [
        ("Blog Part 3", todate(2018, 12, 6)),
        ("Blog Part 2", todate(2018, 12, 5)),
        ("Blog Part 1", todate(2018, 12, 4)),
    ]

    assert actual in (expect, -1)
    johns_blogs = [tuple(r) for r in raw_johns_blogs]
    assert johns_blogs == expected


def run_select_value(conn, queries, db, expect=3):
    # test $
    actual = queries.users.get_count(conn)
    assert actual == expect
    # also with quote escapes
    if _DB[db] in ("mysql", "mariadb"):
        # FIXME does not work
        # actual = queries.misc.my_escape_quotes(conn)
        actual = "L'art du rire"
    else:  # pg, duckdb & sqlite
        actual = queries.misc.escape_quotes(conn)
    assert actual == "L'art du rire"
    # pg-specific check
    if _DB[db] == "postgres":
        actual = queries.misc.pg_escape_quotes(conn)
        assert actual == "'doubled' single quotes"
    # empty result
    none = queries.misc.empty(conn)
    assert none is None


def run_date_time(conn, queries, db):
    if _DB[db] == "sqlite3":
        now = queries.misc.get_now_date_time(conn)
    elif _DB[db] == "duckdb":
        now = queries.misc.duckdb_get_now_date_time(conn)
    elif _DB[db] == "postgres":
        now = queries.misc.pg_get_now_date_time(conn)
    elif _DB[db] in ("mysql", "mariadb"):
        now = queries.misc.my_get_now_date_time(conn)
    else:
        assert False, f"unexpected driver: {db}"
    assert re.match(r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$", now)


#
# Asynchronous tests
#


async def run_async_record_query(conn, queries):
    actual = [dict(r) for r in await queries.users.get_all(conn)]

    assert len(actual) == 3
    assert actual[0] == {
        "userid": 1,
        "username": "bobsmith",
        "firstname": "Bob",
        "lastname": "Smith",
    }


async def run_async_parameterized_query(conn, queries, todate):
    actual = await queries.blogs.get_user_blogs(conn, userid=1)
    expected = [
        ("How to make a pie.", todate(2018, 11, 23)),
        ("What I did Today", todate(2017, 7, 28)),
    ]
    assert actual == expected


async def run_async_parameterized_record_query(conn, queries, db, todate):
    fun = (
        queries.blogs.pg_get_blogs_published_after
        if _DB[db] == "postgres"
        else queries.blogs.sqlite_get_blogs_published_after if _DB[db] == "sqlite3" else None
    )
    records = await fun(conn, published=todate(2018, 1, 1))

    actual = [dict(rec) for rec in records]

    expected = [
        {
            "title": "How to make a pie.",
            "username": "bobsmith",
            "published": "2018-11-23 00:00",
        },
        {"title": "Testing", "username": "janedoe", "published": "2018-01-01 00:00"},
    ]

    assert actual == expected


async def run_async_record_class_query(conn, queries, todate):
    actual = await queries.blogs.get_user_blogs(conn, userid=1)

    expected = [
        UserBlogSummary(title="How to make a pie.", published=todate(2018, 11, 23)),
        UserBlogSummary(title="What I did Today", published=todate(2017, 7, 28)),
    ]

    assert all(isinstance(row, UserBlogSummary) for row in actual)
    assert actual == expected

    one = await queries.blogs.get_latest_user_blog(conn, userid=1)
    assert one == UserBlogSummary(title="How to make a pie.", published=todate(2018, 11, 23))


async def run_async_select_cursor_context_manager(conn, queries, todate):
    async with queries.blogs.get_user_blogs_cursor(conn, userid=1) as cursor:
        actual = [tuple(rec) async for rec in cursor]
        expected = [
            ("How to make a pie.", todate(2018, 11, 23)),
            ("What I did Today", todate(2017, 7, 28)),
        ]
        assert actual == expected


async def run_async_select_one(conn, queries):
    actual = await queries.users.get_by_username(conn, username="johndoe")
    expected = (2, "johndoe", "John", "Doe")
    assert actual == expected


async def run_async_select_value(conn, queries):
    actual = await queries.users.get_count(conn)
    expected = 3
    assert actual == expected


async def run_async_insert_returning(conn, queries, db, todate):
    is_pg = _DB[db] == "postgres"

    fun = queries.blogs.pg_publish_blog if is_pg else queries.blogs.publish_blog

    blogid = await fun(
        conn,
        userid=2,
        title="My first blog",
        contents="Hello, World!",
        published=todate(2018, 12, 4),
    )

    if is_pg:
        blogid, title = blogid
    else:
        blogid, title = blogid, "My first blog"

    if is_pg:
        query = "select blogid, title from blogs where blogid = $1;"
        actual = tuple(
            await conn.fetchrow(
                query,
                blogid,
            )
        )
    else:
        query = "select blogid, title from blogs where blogid = :blogid;"
        async with conn.execute(query, {"blogid": blogid}) as cur:
            actual = await cur.fetchone()
    assert actual == (blogid, title)


async def run_async_delete(conn, queries):
    # Removing the "janedoe" blog titled "Testing"
    actual = await queries.blogs.remove_blog(conn, blogid=2)
    # log.warning(f"actual = {actual}")
    # FIXME all implementations should return the same!
    assert actual == "DELETE 1" or actual == 1

    janes_blogs = await queries.blogs.get_user_blogs(conn, userid=3)
    assert len(janes_blogs) == 0


async def run_async_insert_many(conn, queries, todate):
    blogs = [
        {
            "userid": 2,
            "title": "Blog Part 1",
            "contents": "content - 1",
            "published": todate(2018, 12, 4),
        },
        {
            "userid": 2,
            "title": "Blog Part 2",
            "contents": "content - 2",
            "published": todate(2018, 12, 5),
        },
        {
            "userid": 2,
            "title": "Blog Part 3",
            "contents": "content - 3",
            "published": todate(2018, 12, 6),
        },
    ]
    actual = await queries.blogs.pg_bulk_publish(conn, blogs)
    assert actual is None

    johns_blogs = await queries.blogs.get_user_blogs(conn, userid=2)
    assert johns_blogs == [
        ("Blog Part 3", todate(2018, 12, 6)),
        ("Blog Part 2", todate(2018, 12, 5)),
        ("Blog Part 1", todate(2018, 12, 4)),
    ]


async def run_async_methods(conn, queries):
    users, sorted_users = await asyncio.gather(
        queries.users.get_all(conn), queries.users.get_all_sorted(conn)
    )

    assert [dict(u) for u in users] == [
        {"userid": 1, "username": "bobsmith", "firstname": "Bob", "lastname": "Smith"},
        {"userid": 2, "username": "johndoe", "firstname": "John", "lastname": "Doe"},
        {"userid": 3, "username": "janedoe", "firstname": "Jane", "lastname": "Doe"},
    ]
    assert [dict(u) for u in sorted_users] == [
        {"userid": 1, "username": "bobsmith", "firstname": "Bob", "lastname": "Smith"},
        {"userid": 3, "username": "janedoe", "firstname": "Jane", "lastname": "Doe"},
        {"userid": 2, "username": "johndoe", "firstname": "John", "lastname": "Doe"},
    ]



import inspect
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    AsyncContextManager,
    Callable,
    ContextManager,
    Dict,
    List,
    NamedTuple,
    Optional,
    Union,
    Tuple,
)

try:
    # Python 3.8 (PEP 544)
    from typing import Protocol
except ImportError:  # pragma: no cover
    from typing_extensions import Protocol  # type: ignore


class SQLOperationType(Enum):
    """Enumeration of aiosql operation types."""

    INSERT_RETURNING = 0
    INSERT_UPDATE_DELETE = 1
    INSERT_UPDATE_DELETE_MANY = 2
    SCRIPT = 3
    SELECT = 4
    SELECT_ONE = 5
    SELECT_VALUE = 6


class QueryDatum(NamedTuple):
    query_name: str
    doc_comments: str
    operation_type: SQLOperationType
    sql: str
    record_class: Any
    signature: Optional[inspect.Signature]
    floc: Tuple[Union[Path, str], int]


class QueryFn(Protocol):
    __name__: str
    __signature__: Optional[inspect.Signature]
    sql: str
    operation: SQLOperationType

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...  # pragma: no cover


# Can't make this a recursive type in terms of itself
# QueryDataTree = Dict[str, Union[QueryDatum, 'QueryDataTree']]
QueryDataTree = Dict[str, Union[QueryDatum, Dict]]


class SyncDriverAdapterProtocol(Protocol):
    def process_sql(
        self, query_name: str, op_type: SQLOperationType, sql: str
    ) -> str: ...  # pragma: no cover

    def select(
        self,
        conn: Any,
        query_name: str,
        sql: str,
        parameters: Union[List, Dict],
        record_class: Optional[Callable],
    ) -> List: ...  # pragma: no cover

    def select_one(
        self,
        conn: Any,
        query_name: str,
        sql: str,
        parameters: Union[List, Dict],
        record_class: Optional[Callable],
    ) -> Optional[Any]: ...  # pragma: no cover

    def select_value(
        self, conn: Any, query_name: str, sql: str, parameters: Union[List, Dict]
    ) -> Optional[Any]: ...  # pragma: no cover

    def select_cursor(
        self, conn: Any, query_name: str, sql: str, parameters: Union[List, Dict]
    ) -> ContextManager[Any]: ...  # pragma: no cover

    # TODO: Next major version introduce a return? Optional return?
    def insert_update_delete(
        self, conn: Any, query_name: str, sql: str, parameters: Union[List, Dict]
    ) -> int: ...  # pragma: no cover

    # TODO: Next major version introduce a return? Optional return?
    def insert_update_delete_many(
        self, conn: Any, query_name: str, sql: str, parameters: Union[List, Dict]
    ) -> int: ...  # pragma: no cover

    def insert_returning(
        self, conn: Any, query_name: str, sql: str, parameters: Union[List, Dict]
    ) -> Optional[Any]: ...  # pragma: no cover

    def execute_script(self, conn: Any, sql: str) -> str: ...  # pragma: no cover


class AsyncDriverAdapterProtocol(Protocol):
    def process_sql(
        self, query_name: str, op_type: SQLOperationType, sql: str
    ) -> str: ...  # pragma: no cover

    async def select(
        self,
        conn: Any,
        query_name: str,
        sql: str,
        parameters: Union[List, Dict],
        record_class: Optional[Callable],
    ) -> List: ...  # pragma: no cover

    async def select_one(
        self,
        conn: Any,
        query_name: str,
        sql: str,
        parameters: Union[List, Dict],
        record_class: Optional[Callable],
    ) -> Optional[Any]: ...  # pragma: no cover

    async def select_value(
        self, conn: Any, query_name: str, sql: str, parameters: Union[List, Dict]
    ) -> Optional[Any]: ...  # pragma: no cover

    async def select_cursor(
        self, conn: Any, query_name: str, sql: str, parameters: Union[List, Dict]
    ) -> AsyncContextManager[Any]: ...  # pragma: no cover

    # TODO: Next major version introduce a return? Optional return?
    async def insert_update_delete(
        self, conn: Any, query_name: str, sql: str, parameters: Union[List, Dict]
    ) -> None: ...  # pragma: no cover

    # TODO: Next major version introduce a return? Optional return?
    async def insert_update_delete_many(
        self, conn: Any, query_name: str, sql: str, parameters: Union[List, Dict]
    ) -> None: ...  # pragma: no cover

    async def insert_returning(
        self, conn: Any, query_name: str, sql: str, parameters: Union[List, Dict]
    ) -> Optional[Any]: ...  # pragma: no cover

    async def execute_script(self, conn: Any, sql: str) -> str: ...  # pragma: no cover


DriverAdapterProtocol = Union[SyncDriverAdapterProtocol, AsyncDriverAdapterProtocol]
